#include <PalmOS.h>
#include <VfsMgr.h>
#include <WinGlue.h>
#include "vfsCopyMove.h"
#include "progress.h"
#include "send.h"

//this nonsense is needed to avoid strings ending up in ".data" somehow. that wastes dynamic ram



static const char mStrInternalToInternalCopy[] = "Internal memory to internal memory ops make no sense";
static const char mStrTaskFailedSuccessfully[] = "Task failed successfully";
static const char mStrToPerformThisAction[] = " to perform this action";
static const char mStrMustBe4charsLong[] = " must be precisely 4 characters long";
static const char mStrAfileToCopyOrCut[] = "a file to copy/cut";
static const char mStrAnErrorOccured[] = "An error occured: ";
static const char mStrExpansionCard[] = "an expansion card";
static const char mStrNeedToSelect[] = "You need to select ";
static const char mStrDirSeparator[] = "/";
static const char mStrIsRequired[] = " is required for this functionality";
static const char mStrDatabase[] = "Database ";
static const char mStrPalmOS30[] = "PalmOS 3.0 or later";
static const char mStrInternal[] = "Internal";
static const char mStrOccured[] = " occured";
static const char mStrError0x[] = "Error 0x";
static const char mStrCardNo[] = "Card #";
static const char mStrAnItem[] = "an item";
static const char mStrAfile[] = "a file";
static const char mStrSpace[] = " ";
static const char mStrSorry[] = "Sorry, ";
static const char mStrAdir[] = "a directory";
static const char mStrType[] = "type";
static const char mStrCrid[] = "creator ID";
static const char mStrDot[] = ".";
static const char mStrPdb[] = ".pdb";
static const char mStrPrc[] = ".prc";






#define VOL_NO_INTERNAL_WITH_ROM			(-1)
#define VOL_NO_INTERNAL_NO_ROM				(-2)
#define MAX_VOLUMES							10

#define TYPE_INTERNAL_RAM					0x00
#define TYPE_INTERNAL_ROM					0x01
#define TYPE_VFS_FILE						0x02
#define TYPE_VFS_DIR						0x03

#define EVENT_TYPE_VOLUMES_CHANGED			(firstUserEvent + 0)

struct FileEntry {
	struct FileEntry *next;
	UInt32 size;
	UInt8 type		:	2;
	UInt8 cardNo	:	2;
	char name[];
};

struct VolumeEntry {
	UInt16 vrn;
	char name[16];
};

struct PreciseFileSpecifier {
	UInt8 type		:2;
	UInt8 isMove	:1;
	union {
		struct {
			UInt8 cardNo;
			char name[];
		} internal;
		struct {
			UInt16 vrn;
			char fullPath[];
		} external;
	};
};



typedef Int8 (*SortCompareF)(const struct FileEntry *lhs, struct FileEntry *rhs);


static UInt8 mNumVolumes;
static struct VolumeEntry mVolumes[MAX_VOLUMES];
static Int8 mCurVolume = VOL_NO_INTERNAL_WITH_ROM;
static MemHandle mCurPath;
static struct FileEntry *mFiles = NULL, *mCurDetailsFile;
static SortCompareF mCurCompareF;
static struct PreciseFileSpecifier *mPickedFile;

static void uiPrvErrAlert(Err e)
{
	if (e == errNone)
		SysFatalAlert(mStrTaskFailedSuccessfully);
	else {
	
		UInt8 specifier = e & 0xff;
		MemHandle mh;
		char x[9];
		
		if (specifier) {
		
			mh = DmGetResource('err!', 0x1000 + (e >> 8));
			if (mh) {
				
				char *res = MemHandleLock(mh), *end = res + MemHandleSize(mh), *cls = res;
				
				while (res < end && specifier--)
					res += StrLen(res) + 1;
				
				if (res < end) {
					
					FrmCustomAlert(1000, mStrAnErrorOccured, cls, res);
					MemHandleUnlock(mh);
					return;
				}
				
				MemHandleUnlock(mh);
			}
		}
		
		StrIToH(x, e);
		FrmCustomAlert(1000, mStrError0x, x + 4, mStrOccured);
	}
}

static void uiPrvFormVfsNameInBuffer(char *ret, const char *first, const char *last)
{
	UInt16 firstLen;
	
	firstLen = StrLen(first);
		
	StrCopy(ret, first);
	if (ret[firstLen - 1] == '/')
		firstLen--;
	else
		ret[firstLen] = '/';
	StrCopy(ret + firstLen + 1, last);
}

static char *uiPrvFormVfsNameFromParts(const char *first, const char *last)
{
	char *ret = MemPtrNew(StrLen(first) + StrLen(last) + 2);
	
	if (ret)
		uiPrvFormVfsNameInBuffer(ret, first ,last);
	
	return ret;
}

static char* uiPrvFormVfsName(const char *last)
{
	char *ret = uiPrvFormVfsNameFromParts(MemHandleLock(mCurPath), last);
	MemHandleUnlock(mCurPath);
	return ret;
}

static void* uiPrvFrmGetObjectPtrFromId(FormPtr fp, UInt16 id)
{
	return FrmGetObjectPtr(fp, FrmGetObjectIndex(fp, id));
}

static void uiPrvSetTextField(FormPtr fp, UInt16 fldIdx, const char *strIn, UInt16 len)
{
	FieldPtr fi = uiPrvFrmGetObjectPtrFromId(fp, fldIdx);
	MemHandle mh;
	
	mh = FldGetTextHandle(fi);
	FldSetTextHandle(fi, NULL);
	if (mh)
		MemHandleFree(mh);
	
	mh = MemHandleNew(len + 1);
	if (mh) {
		char *dst = MemHandleLock(mh);
		
		MemMove(dst, strIn, len);
		dst[len] = 0;
		
		MemHandleUnlock(mh);
		FldSetTextHandle(fi, mh);
	}
}

static const char* uiPrvGetFieldText(FormPtr fp, UInt16 fldIdx)
{
	return FldGetTextPtr(uiPrvFrmGetObjectPtrFromId(fp, fldIdx));
}

static UInt16 filesPrvCount(const struct FileEntry *list)
{
	const struct FileEntry *t;
	UInt16 count = 0;
	
	for (t = list, count = 0; t; t = t->next, count++);
	
	return count;
}

static struct FileEntry* filesPrvSort(struct FileEntry *list, SortCompareF cmpF)
{
	struct FileEntry *t, *p, *ret = NULL;
	UInt16 count = 0;

	count = filesPrvCount(list);
	
	//if zero or one, we're done
	if (count < 2)
		return list;
	
	//get middle of list
	for (t = list, count /= 2; count; p = t, t = t->next, count--);
	
	//terminate first list
	p->next = NULL;
	
	//sort each half
	list = filesPrvSort(list, cmpF);
	t = filesPrvSort(t, cmpF);
	
	//merge
	ret = NULL;
	p = NULL;
	while (list || t) {
		
		struct FileEntry *lesser;
		
		if (!t || (list && cmpF(list, t) <= 0)) {
			
			lesser = list;
			list = list->next;
		}
		else {
			
			lesser = t;
			t = t->next;
		}
		
		
		lesser->next = NULL;
		if (p)
			p->next = lesser;
		else
			ret = lesser;
		p = lesser;
	}
	
	return ret;
}

static void filesPrvFree(struct FileEntry *list)
{
	struct FileEntry *cur, *next;
	
	for (cur = list; cur; cur = next) {
		
		next = cur->next;
		MemPtrFree(cur);
	}
}

static UInt32 getLocalIDsize(UInt16 cardNo, LocalID LID)
{
	void* chunk;
	UInt32 ret;
	
	if (!LID)
		return 0;
	
	chunk = MemLocalIDToLockedPtr(LID, cardNo);
	ret = MemPtrSize(chunk);
	MemPtrUnlock(chunk);
	
	return ret;
}

static UInt32 estimateDbSize(UInt16 cardNo, LocalID LID, Boolean isResDb)
{
	UInt32 size;
	
	if (errNone != (DmDatabaseSize(cardNo, LID, NULL, &size, NULL)))
		size = 0;
	
	return size;
}

static struct FileEntry* filesPrvEnumerateInternal(Boolean includeRom)
{
	struct FileEntry *list = NULL, *t;
	char name[dmDBNameLength + 1];
	UInt16 card, numDbs, i;
			
	name[sizeof(name - 1)] = 0;
	for (card = 0; card< MemNumCards(); card++) {
		
		
		numDbs = DmNumDatabases(card);
		for (i = 0; i < numDbs; i++) {
			
			LocalID LID = DmGetDatabase(card, i);
			UInt16 attrs;
			
			if (LID && errNone == DmDatabaseInfo(card, LID, name, &attrs, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
				
				Boolean isRom = (MemLocalIDKind(LID) != memIDHandle);
					
				if (!includeRom && isRom)
					continue;
				
				t = MemPtrNew(sizeof(*t) + StrLen(name) + 1);
				if (t) {
					
					t->type = isRom ? TYPE_INTERNAL_ROM : TYPE_INTERNAL_RAM;
					t->cardNo = card;
					t->size = estimateDbSize(card, LID, attrs & dmHdrAttrResDB);
					StrCopy(t->name, name);
					
					t->next = list;
					list = t;
				}
			}
		}
	}
	return list;
}

static struct FileEntry* filesPrvEnumerateVfs(UInt16 vrn, const char *path)
{
	struct FileEntry *list = NULL, *t;
	UInt32 iter = vfsIteratorStart;
	FileRef dir, file;
	FileInfoType fi;
	char *fullPath;
	
	fi.nameP = MemPtrNew((fi.nameBufLen = 255) + 1);
	if (!fi.nameP)
		return NULL;
	
	fullPath = MemPtrNew(StrLen(path) + fi.nameBufLen + 1);
	
	if (errNone == VFSFileOpen(vrn, path, vfsModeRead, &dir) && dir) {
		
		while (iter != vfsIteratorStop && errNone == VFSDirEntryEnumerate(dir, &iter, &fi)) {
			
			t = MemPtrNew(sizeof(*t) + StrLen(fi.nameP) + 1);
			if (t) {
				
				t->type = (fi.attributes & vfsFileAttrDirectory) ? TYPE_VFS_DIR : TYPE_VFS_FILE;
				t->size = 0;
				StrCopy(t->name, fi.nameP);
				
				t->next = list;
				list = t;
				
				if (t->type == TYPE_VFS_FILE && fullPath) {
					
					uiPrvFormVfsNameInBuffer(fullPath, path, fi.nameP);
					
					if (errNone == VFSFileOpen(vrn, fullPath, vfsModeRead, &file) && file) {
						
						if (errNone != VFSFileSize(file, &t->size))
							t->size = 0;
						
						VFSFileClose(file);
					}
				}
			}
		}
		
		VFSFileClose(dir);
	}
	
	if (fullPath)
		MemPtrFree(fullPath);
	
	MemPtrFree(fi.nameP);
	return list;
}

static struct FileEntry* filesPrvEnumerate(void)
{
	if (mCurVolume < 0) {
	
		return filesPrvEnumerateInternal(mCurVolume == VOL_NO_INTERNAL_WITH_ROM);
	}
	else {
		
		struct FileEntry *ret;
		
		ret = filesPrvEnumerateVfs(mVolumes[mCurVolume].vrn, MemHandleLock(mCurPath));
		MemHandleUnlock(mCurPath);
		
		return ret;
	}
}

static Int8 uiPrvSortCompareByName(const struct FileEntry *lhs, struct FileEntry *rhs)
{
	return StrCaselessCompare(lhs->name, rhs->name);
}

static Int8 uiPrvSortCompareByNameReverse(const struct FileEntry *lhs, struct FileEntry *rhs)
{
	return -StrCaselessCompare(lhs->name, rhs->name);
}

static Int8 uiPrvSortCompareBySize(const struct FileEntry *lhs, struct FileEntry *rhs)
{
	return (lhs->size > rhs->size) ? 1 : ((lhs->size == rhs->size) ? 0 : -1);
}

static Int8 uiPrvSortCompareBySizeReverse(const struct FileEntry *lhs, struct FileEntry *rhs)
{
	return (lhs->size > rhs->size) ? -1 : ((lhs->size == rhs->size) ? 0 : 1);
}

static void uiPrvObjectHide(FormPtr fp, UInt16 objId)
{
	UInt16 idx = FrmGetObjectIndex(fp, objId);
	Boolean needWorkaroundBitmapHiding;
	UInt32 osVer;
	void *ptr;
	
	/*
		PalmOS 3.1 and below cannot properly hide form bitmaps since they test for "visible" bit but never set it
		This patch sets visible and usable bits before hiding it to force an erase as expected, and clears them after.
		FrmShowObject() works as expected since when visibleis  clear, it will draw the object
	*/
	
	needWorkaroundBitmapHiding = FrmGetObjectType(fp, idx) == frmBitmapObj && (errNone != FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osVer) || osVer < sysMakeROMVersion(3, 5, 0, sysROMStageRelease, 0));
	
	if (needWorkaroundBitmapHiding) {
		ptr = FrmGetObjectPtr(fp, idx);
		(*(UInt8*)ptr) |= 0xa0;
	}
	
	FrmHideObject(fp, idx);
	
	if (needWorkaroundBitmapHiding)
		(*(UInt8*)ptr) &=~ 0xa0;
}

static void uiPrvObjectShow(FormPtr fp, UInt16 objId)
{
	FrmShowObject(fp, FrmGetObjectIndex(fp, objId));
}

static void uiRefresh(void)
{
	FormPtr fp = FrmGetFormPtr(1000);
	ListPtr list = uiPrvFrmGetObjectPtrFromId(fp, 1000);

	LstSetSelection(list, noListSelection);
	LstSetListChoices(list, NULL, filesPrvCount(mFiles));
	
	if (mCurVolume < 0) {	//internal memory
		
		uiPrvObjectHide(fp, 5000);
		uiPrvObjectShow(fp, 4000);
		CtlSetValue(uiPrvFrmGetObjectPtrFromId(fp, 4000), mCurVolume == VOL_NO_INTERNAL_WITH_ROM);
		
		uiPrvObjectHide(fp, 7000);
		uiPrvObjectHide(fp, 7001);
		uiPrvObjectHide(fp, 1003);
		uiPrvObjectHide(fp, 1004);
	}
	else {
		
		const char *curPath = MemHandleLock(mCurPath);
		UInt16 curPathLen = StrLen(curPath);
		
		uiPrvObjectHide(fp, 4000);
		uiPrvObjectShow(fp, 5000);
		
		uiPrvSetTextField(fp, 5000, curPath, curPathLen);
		
		if (curPathLen == 1 && curPath[0] == '/') {	//is root?
			
			uiPrvObjectHide(fp, 7001);
			uiPrvObjectHide(fp, 1004);
		}
		else {
		
			uiPrvObjectShow(fp, 7001);
			uiPrvObjectShow(fp, 1004);
		}
		
		uiPrvObjectShow(fp, 1003);
		uiPrvObjectShow(fp, 7000);
		
		MemHandleUnlock(mCurPath);
	}
	
	if (mPickedFile)
		uiPrvObjectShow(fp, 1005);
	else
		uiPrvObjectHide(fp, 1005);
	
	FrmDrawForm(fp);
}

static void uiPrvFileListResortAndUiRefresh(void)
{
	mFiles = filesPrvSort(mFiles, mCurCompareF);
	
	uiRefresh();
}

static void uiPrvFileListAndUiRefresh(void)
{
	filesPrvFree(mFiles);
	mFiles = filesPrvEnumerate();
	
	uiPrvFileListResortAndUiRefresh();
}

static Boolean uiPrvVolumesChanged(void)	//return true if redraw is needed
{
	Boolean redrawNeeded = false, pickedFileAffected = mPickedFile && (mPickedFile->type == TYPE_VFS_FILE || mPickedFile->type == TYPE_VFS_DIR);
	UInt16 curVrn;
	UInt32 dummy;
	UInt8 idx = 0;
	
	if (mCurVolume >= 0)
		curVrn = mVolumes[mCurVolume].vrn;
	
	if (errNone == FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &dummy) && dummy >= 1) {
		
		Boolean curFound = false, pickedFileFound = false;
		UInt16 vrn;
		
		for (dummy = vfsIteratorStart; dummy != vfsIteratorStop && errNone == VFSVolumeEnumerate(&vrn, &dummy); idx++) {
			
			mVolumes[idx].vrn = vrn;
			
			if (pickedFileAffected && mPickedFile->external.vrn == vrn)
				pickedFileFound = true;

			if (mCurVolume >= 0 && curVrn == vrn) {
				curFound = true;
				mCurVolume = idx;
			}
			
			if (errNone != VFSVolumeGetLabel(vrn, mVolumes[idx].name, sizeof(mVolumes[idx].name)) || !mVolumes[idx].name[0]) {
				
				StrCopy(mVolumes[idx].name, mStrCardNo);
				mVolumes[idx].name[5] = idx + '0';
			}
		}
		mNumVolumes = idx;
		if (mCurVolume >= 0 && !curFound) {
			redrawNeeded = true;
			mCurVolume = VOL_NO_INTERNAL_WITH_ROM;
		}
			
		if (pickedFileAffected && !pickedFileFound) {
			
			MemPtrFree(mPickedFile);
			mPickedFile = NULL;
			
			if (!redrawNeeded)	//mild ui redraw
				uiRefresh();///aaa
		}
	}
	
	LstSetListChoices(uiPrvFrmGetObjectPtrFromId(FrmGetFormPtr(1000), 3100), NULL, idx + 1);

	return redrawNeeded;
}

static Err uiPrvVolumesChangedNotif(SysNotifyParamType *notifyParamsP)
{
	EventType e;
	
	MemSet(&e, sizeof(e), 0);
	e.eType = EVENT_TYPE_VOLUMES_CHANGED;
	
	EvtAddEventToQueue(&e);
	
	return errNone;
}

static void uiPrvSizeToString(char *dst, UInt32 size)
{
	char unit;
	UInt32 div;
	UInt16 frac;

	if (size < 1024) {
		div = 1;
		unit = 0;
	}
	else if (size < 900 * 1024UL) {
		
		div = 1024UL;
		unit = 'K';
	}
	else if (size < 900 * 1024UL * 1024UL) {
		
		div = 1024UL * 1024UL;
		unit = 'M';
	}
	else {
		
		div = 1024UL * 1024UL * 1024UL;
		unit = 'G';
	}
	StrIToA(dst, size / div);
	if (frac = size % div) {
		
		StrCat(dst, mStrDot);
		StrIToA(dst + StrLen(dst), frac / 100);
	}
	dst += StrLen(dst);
	if (unit) {
		*dst++ = ' ';
		*dst++ = unit;
		*dst++ = 0;
	}
}

static void uiPrvFileListDrawFunc(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
	//slow, but saves memory
	const struct FileEntry *t = mFiles;
	char szStr[12];
	
	if (itemNum < 0)
		return;
	
	while (itemNum-- && t)
		t = t->next;
	
	WinGlueDrawTruncChars(t->name, StrLen(t->name), bounds->topLeft.x, bounds->topLeft.y, 122);
	if (t->type == TYPE_VFS_DIR) {
		
		MemHandle mh = DmGetResource('Tbmp', 1000);
		WinDrawBitmap(MemHandleLock(mh), bounds->topLeft.x + 123, bounds->topLeft.y);
		MemHandleUnlock(mh);
		DmReleaseResource(mh);
	}
	else {
	
		uiPrvSizeToString(szStr, t->size);
		WinGlueDrawTruncChars(szStr, StrLen(szStr), bounds->topLeft.x + 123, bounds->topLeft.y, 35);
	}
}

static const char* uiPrvVolumeListNameByIndex(Int16 idx)
{
	return idx < 1 ? mStrInternal : mVolumes[idx - 1].name;
}

static void uiPrvVolumeListDrawFunc(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
	const char *name = uiPrvVolumeListNameByIndex(itemNum);
	Int16 x = bounds->topLeft.x;
	
	if (itemNum) {
		MemHandle mh = DmGetResource('Tbmp', 1001);
		WinDrawBitmap(MemHandleLock(mh), x, bounds->topLeft.y + 1);
		MemHandleUnlock(mh);
		DmReleaseResource(mh);
		x += 7;
	}

	WinDrawChars(name, StrLen(name), x, bounds->topLeft.y);
}

static void uiPrvLackOfFeatureError(const char *str)
{
	FrmCustomAlert(1000, mStrSorry, str, mStrIsRequired);
}

static Boolean uiPrvVerifyFeature(UInt32 ftrName, UInt16 ftrVal, UInt32 minVal, const char *str)
{
	UInt32 ver;
	
	if (errNone == FtrGet(ftrName, ftrVal, &ver) && ver >= minVal)
		return true;
	
	uiPrvLackOfFeatureError(str);
	return false;
}

static Boolean uiPrvVerifyOs(UInt32 minVal, const char *str)
{
	return uiPrvVerifyFeature(sysFtrCreator, sysFtrNumROMVersion, minVal, str);
}

static void uiPrvItemSend(struct FileEntry *f)
{
	char *path;
	Err e;
	
	if (!uiPrvVerifyOs(sysMakeROMVersion(3, 0, 0, sysROMStageRelease, 0), mStrPalmOS30))
		return;

	switch (f->type) {
		case TYPE_INTERNAL_RAM:
		case TYPE_INTERNAL_ROM:
			e = sendPrvInternal(f->cardNo, f->name, f->size);
			break;
		
		case TYPE_VFS_FILE:
			path = uiPrvFormVfsName(f->name);
			if (!path) {
				e = memErrNotEnoughSpace;
				break;
			}
			e = sendPrvFromCard(mVolumes[mCurVolume].vrn, f->name, path);
			MemPtrFree(path);
			break;
		
		default:
			FrmCustomAlert(1000, mStrNeedToSelect, mStrAfile, mStrToPerformThisAction);
			return;
	}
	
	if (e != errNone && e != exgErrUserCancel)
		uiPrvErrAlert(e);
}


static void uiPrvItemCopy(struct FileEntry *f, Boolean forMove)
{
	if (!mNumVolumes)
		uiPrvLackOfFeatureError(mStrExpansionCard);
	
	if (forMove && f->type == TYPE_INTERNAL_ROM) {
	
		uiPrvErrAlert(dmErrROMBased);
		return;
	}
	
	if (f->type == TYPE_INTERNAL_ROM || f->type == TYPE_INTERNAL_RAM) {
		mPickedFile = MemPtrNew(sizeof(*mPickedFile) + StrLen(f->name) + 1);
		if (mPickedFile) {
			mPickedFile->internal.cardNo = f->cardNo;
			StrCopy(mPickedFile->internal.name, f->name);
		}
	}
	else {
		mPickedFile = MemPtrNew(sizeof(*mPickedFile) + StrLen(f->name) + StrLen(MemHandleLock(mCurPath)) + 2);
		MemHandleUnlock(mCurPath);
		
		if (mPickedFile) {
			
			mPickedFile->external.vrn = mVolumes[mCurVolume].vrn;
			uiPrvFormVfsNameInBuffer(mPickedFile->external.fullPath, MemHandleLock(mCurPath), f->name);
			MemHandleUnlock(mCurPath);
		}
	}
	
	if (!mPickedFile) {
		uiPrvErrAlert(memErrNotEnoughSpace);
		return;
	}
		
	mPickedFile->type = f->type;
	mPickedFile->isMove = forMove;
	uiRefresh();
}

static Err uiPrvItemProgress(UInt32 totalBytes, UInt32 offset, void *userDataP)
{
	struct Progress *progress = (struct Progress*)userDataP;
	
	EvtResetAutoOffTimer();
	progressUpdate(progress, offset, totalBytes);
	return errNone;
}

//VFSExportDatabaseToFileCustom has a bug in PalmOS versions below 5.0, it does not initialize its internal "offset" member and thus
// the progress reports we get are offset by a random value that wa on stack.  It gets worse. The buffering setup employed by
// VFSExportDatabaseToFile*() causes the offset to never be updated. So first you'll be told you're 0 bytes into the file, and then
// you'll be told you're $RANDOM_UINT32 bytesin, and will keep being told this till done...
// We DO know that the very first non-zero write is precisely 72 bytes in size. We can use that to keep track of our offset.
// To work around the buffering issue, we need to disable the buffering. There is no option to do that officially, but patching
// DmNewHandle(NULL, ...) to reject allocations that are 32K 16K, 8K .. 256 bytes works well

struct WorkaroundState {
	MemHandle (*DmNewHandleOldTrap)(DmOpenRef db, UInt32 size);
	VFSExportProcPtr exportProcF;
	void *exportProcD;
	UInt32 initialOffset;
	UInt16 expectedAllocSz;
};

static MemHandle pDmNewHandleC(struct WorkaroundState *was, void *calledFrom, DmOpenRef db, UInt32 size)
{
	if (!db && size == (UInt32)was->expectedAllocSz) {
		was->expectedAllocSz >>= 1;
		return NULL;
	}
	
	return was->DmNewHandleOldTrap(db, size);
}

asm static MemHandle pDmNewHandleAsm(DmOpenRef db, UInt32 size)
{
	move.l #0x12345678, -(sp)
	jsr    pDmNewHandleC
	addq   #4, sp
	rts
}

static Err MyVFSExportDatabaseToFileCustomProgressF(UInt32 totalBytes, UInt32 offset, void *userDataP)
{
	struct WorkaroundState *was = (struct WorkaroundState*)userDataP;
	
	if (was->expectedAllocSz == 128)			//first call after we denied the llocations. offset must be ZERO
		was->expectedAllocSz = 64;
	else if (was->expectedAllocSz == 64) {		//second call - 72 bytes have been written
		
		was->initialOffset = offset - 72;
		was->expectedAllocSz = 32;
		offset = 72;
	}
	else if (was->expectedAllocSz == 32) {
		
		offset -= was->initialOffset;
	}
	
	return was->exportProcF(totalBytes, offset, was->exportProcD);
}

static Err MyVFSExportDatabaseToFileCustom(UInt16 volRefNum, const Char *pathNameP, UInt16 cardNo, LocalID dbID, VFSExportProcPtr exportProcP, void *userDataP)
{
	UInt32 romVersion;
	
	if (errNone != FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion) || romVersion >= sysMakeROMVersion(5, 0, 0, sysROMStageDevelopment, 0))
		return VFSExportDatabaseToFileCustom(volRefNum, pathNameP, cardNo, dbID, exportProcP, userDataP);
	else {	//workaround
		
		struct WorkaroundState was;
		Err ret;
		
		was.DmNewHandleOldTrap = SysGetTrapAddress(sysTrapDmNewHandle);
		MemSemaphoreReserve(true);
		*(void**)(((UInt16*)pDmNewHandleAsm) + 1) = &was;
		MemSemaphoreRelease(true);
		SysSetTrapAddress(sysTrapDmNewHandle, pDmNewHandleAsm);
		was.expectedAllocSz = 0x8000;
		was.exportProcF = exportProcP;
		was.exportProcD = userDataP;
		
		ret = VFSExportDatabaseToFileCustom(volRefNum, pathNameP, cardNo, dbID, exportProcP, userDataP);
		
		SysSetTrapAddress(sysTrapDmNewHandle, was.DmNewHandleOldTrap);
		
		return ret;
	}
}


static void uiPrvItemPaste(void)
{
	Boolean refreshNeeded = false;
	struct Progress *progress;
	Err e = errNone;
	char *fullPath;
	LocalID LID;
	UInt16 i;
	
	if (!mNumVolumes)
		uiPrvLackOfFeatureError(mStrExpansionCard);
	else if (!mPickedFile)
		FrmCustomAlert(1000, mStrNeedToSelect, mStrAfileToCopyOrCut, mStrToPerformThisAction);
	else if (mPickedFile->type == TYPE_INTERNAL_ROM || mPickedFile->type == TYPE_INTERNAL_RAM) {
		
		if (mCurVolume < 0)
			FrmCustomAlert(1000, mStrInternalToInternalCopy, mStrSpace, mStrSpace);
		else  {
			
			char name[dmDBNameLength + 5], ch;
			UInt16 attr;
			
			LID = DmFindDatabase(mPickedFile->internal.cardNo, mPickedFile->internal.name);
			if (!LID) {
				e = dmErrCantFind;
				goto out;
			}
			
			//create a filename
			(void)DmDatabaseInfo(mPickedFile->internal.cardNo,LID, NULL, &attr, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
			for (i = 0; i < dmDBNameLength && (ch = mPickedFile->internal.name[i]) != 0; i++) {
				
				if (ch < 0x20 || ch >= 0x7f)
					ch = '_';
				else switch (ch) {
					case '&':
					case '/':
					case '\\':
					case ':':
					case '*':
					case '"':
					case '<':
					case '>':
					case '|':
						ch = '_';
						break;
				}
				
				name[i] = ch;
			}
			StrCopy(name + i, (attr & dmHdrAttrResDB) ? mStrPrc : mStrPdb);
			
			fullPath = uiPrvFormVfsName(name);
			if (!fullPath) {
				
				e = memErrNotEnoughSpace;
				goto out;
			}
			
			progress = progressStart(mPickedFile->internal.name);
			e = MyVFSExportDatabaseToFileCustom(mVolumes[mCurVolume].vrn, fullPath, mPickedFile->internal.cardNo, DmFindDatabase(mPickedFile->internal.cardNo, mPickedFile->internal.name), uiPrvItemProgress, progress);
			progressFinish(progress);
			MemPtrFree(fullPath);
			if (e != errNone)
				goto out;
			
			refreshNeeded = true;
			
			if (mPickedFile->isMove)
				e = DmDeleteDatabase(mPickedFile->internal.cardNo, LID);
		}
	}
	else if (mCurVolume < 0) {
		
		if (mPickedFile->type == TYPE_VFS_DIR)
			FrmCustomAlert(1000, mStrNeedToSelect, mStrAfile, mStrToPerformThisAction);
		else {
			
			UInt16 cardNo;
		
			progress = progressStart(mPickedFile->external.fullPath);
			e = VFSImportDatabaseFromFileCustom(mPickedFile->external.vrn, mPickedFile->external.fullPath, &cardNo, &LID, uiPrvItemProgress, progress);
			progressFinish(progress);
			if (e != errNone)
				goto out;
			
			refreshNeeded = true;
			
			if (mPickedFile->isMove)
				e = VFSFileDelete(mPickedFile->external.vrn, mPickedFile->external.fullPath);
		}
	}
	else {
		
		e = vfsCopyMove(mPickedFile->external.vrn, mPickedFile->external.fullPath, mVolumes[mCurVolume].vrn, MemHandleLock(mCurPath), mPickedFile->isMove);
		refreshNeeded = true;
		MemHandleUnlock(mCurPath);
	}


out:
	if (e != errNone)
		uiPrvErrAlert(e);
	if (mPickedFile)
		MemPtrFree(mPickedFile);
	mPickedFile = NULL;
	if (refreshNeeded)
		uiPrvFileListAndUiRefresh();
	else
		uiRefresh();	//milder but still redraw the ui
}

static Err uiPrvVfsRecursiveDelete(UInt16 vrn, const char *path)
{
	UInt32 iter = vfsIteratorStart;
	char *fullPath = NULL;
	FileInfoType fit;
	FileRef dir;
	Err e;
	
	fit.nameP = MemPtrNew((fit.nameBufLen = 255) + 1);
	if (!fit.nameP)
		return memErrNotEnoughSpace;
	
	e = VFSFileOpen(vrn, path, vfsModeRead, &dir);
	if (e != errNone)
		goto out;
	
	while (iter != vfsIteratorStop) {
		
		e = VFSDirEntryEnumerate(dir, &iter, &fit);
		if (e == expErrEnumerationEmpty) {
			e = errNone;
			break;
		}
		
		fullPath = uiPrvFormVfsNameFromParts(path, fit.nameP);
		if (!fullPath) {
			e = memErrNotEnoughSpace;
			goto out;
		}
		

		e = (fit.attributes & vfsFileAttrDirectory) ? uiPrvVfsRecursiveDelete(vrn, fullPath) : VFSFileDelete(vrn, fullPath);
		MemPtrFree(fullPath);
		fullPath = NULL;
		
		if (e != errNone)
			goto out;
	}
	
	e = VFSFileDelete(vrn, path);
	
out:
	MemPtrFree(fit.nameP);
	if (fullPath)
		MemPtrFree(fullPath);
	return e;
}

static void uiPrvItemDelete(struct FileEntry *f)
{
	Err e = errNone;
	LocalID LID;
	char *name;

	switch (f->type) {
		case TYPE_INTERNAL_RAM:
			LID = DmFindDatabase(f->cardNo, f->name);
			if (!LID)
				e = dmErrCantFind;
			else
				e = DmDeleteDatabase(f->cardNo, LID);
			break;
		
		case TYPE_INTERNAL_ROM:
			e = dmErrROMBased;
			break;
		
		case TYPE_VFS_FILE:
			name = uiPrvFormVfsName(f->name);
			if (!name)
				e = memErrNotEnoughSpace;
			else {
			
				e = VFSFileDelete(mVolumes[mCurVolume].vrn, name);
				MemPtrFree(name);
			}
			break;
		
		case TYPE_VFS_DIR:
			name = uiPrvFormVfsName(f->name);
			if (!name)
				e = memErrNotEnoughSpace;
			else {
			
				e = uiPrvVfsRecursiveDelete(mVolumes[mCurVolume].vrn, name);
				MemPtrFree(name);
			}
			break;
		
		default:
			e = sysErrParamErr;
			break;
	}
	
	if (e != errNone)
		uiPrvErrAlert(e);
	else {

		FormPtr fp = FrmGetFormPtr(1000);
		ListPtr list = uiPrvFrmGetObjectPtrFromId(fp, 1000);
		struct FileEntry *prev, *cur;
		
		for (prev = NULL, cur = mFiles; cur != f; prev = cur, cur = cur->next);
		
		if (prev)
			prev->next = f->next;
		else
			mFiles = f->next;
		
		MemPtrFree(f);
		
		LstSetListChoices(list, NULL, LstGetNumberOfItems(list) - 1);
		FrmDrawForm(fp);
	}
}

static void uiPrvPopulateInternalDetailsForm(FormPtr fp)
{
	UInt16 attr, ver, i;
	UInt32 type, crid;
	char verStr[6];

	DmDatabaseInfo(mCurDetailsFile->cardNo, DmFindDatabase(mCurDetailsFile->cardNo, mCurDetailsFile->name), NULL, &attr, &ver, NULL, NULL, NULL, NULL, NULL, NULL, &type, &crid);
	
	StrIToA(verStr, ver);
	uiPrvSetTextField(fp, 3000, mCurDetailsFile->name, StrLen(mCurDetailsFile->name));
	uiPrvSetTextField(fp, 3001, (const char*)&type, 4);
	uiPrvSetTextField(fp, 3002, (const char*)&crid, 4);
	uiPrvSetTextField(fp, 3003, verStr, StrLen(verStr));
	
	for (i = 0; i < 12; i++)
		CtlSetValue(uiPrvFrmGetObjectPtrFromId(fp, 2000 + i), !!(attr & (1 << i)));
	
	if (mCurDetailsFile->type == TYPE_INTERNAL_ROM) {
	
		uiPrvObjectShow(fp, 1002);
		uiPrvObjectHide(fp, 8000);
	}
	else {
	
		uiPrvObjectShow(fp, 8000);
		uiPrvObjectHide(fp, 1002);
	}
}

static Boolean uiPrvPopulateExternalDetailsForm(FormPtr fp)
{
	char *fullPath;
	UInt32 attr;
	FileRef f;
	UInt16 i;
	Err e;
	
	fullPath = uiPrvFormVfsName(mCurDetailsFile->name);
	if (!fullPath)
		return false;
	e = VFSFileOpen(mVolumes[mCurVolume].vrn, fullPath, vfsModeRead, &f);
	MemPtrFree(fullPath);
	
	if (e != errNone) {
		uiPrvErrAlert(e);
		return false;
	}
	
	e = VFSFileGetAttributes(f, &attr);
	if (e != errNone)
		goto out_closefile;
	
	if (attr & vfsFileAttrDirectory) {
	
		uiPrvObjectHide(fp, 9001);
		uiPrvObjectHide(fp, 3001);
	}
	else {
	
		char sizeStr[11];
		UInt32 size;
	
		uiPrvObjectShow(fp, 9001);
		uiPrvObjectShow(fp, 3001);
		
		e = VFSFileSize(f, &size);
		if (e != errNone)
			goto out_closefile;
			
		StrIToA(sizeStr, size);
		uiPrvSetTextField(fp, 3001, sizeStr, StrLen(sizeStr));
	}
	
	uiPrvSetTextField(fp, 3000, mCurDetailsFile->name, StrLen(mCurDetailsFile->name));
	
	for (i = 0; i < 6; i++)
		CtlSetValue(uiPrvFrmGetObjectPtrFromId(fp, 2000 + i), !!(attr & (1UL << i)));

out_closefile:
	VFSFileClose(f);
	
	if (e != errNone) {
		uiPrvErrAlert(e);
		return false;
	}
	
	return true;	
}

static Err uiPrvCurFileRenamed(const char *newName)
{
	if (MemPtrSize(mCurDetailsFile) < sizeof(struct FileEntry) + StrLen(newName) + 1) {	//we need a new chunk - allocate and replace the current

		struct FileEntry *t = MemPtrNew(sizeof(struct FileEntry) + StrLen(newName) + 1);
				
		if (!t)
			return memErrNotEnoughSpace;
		
		MemMove(t, mCurDetailsFile, MemPtrSize(mCurDetailsFile));
		
		//replace it in the file list (yes, we'll re-sort later)
		if (mCurDetailsFile == mFiles)
			mFiles = t;
		else {
			struct FileEntry *prev = NULL, *cur;
			
			for (cur = mFiles; cur && cur != mCurDetailsFile; prev = cur, cur = cur->next);
			
			prev->next = t;
		}
		
		if (mPickedFile && mPickedFile->type == TYPE_INTERNAL_RAM && mPickedFile->internal.cardNo == mCurDetailsFile->cardNo && !StrCompare(mCurDetailsFile->name, mPickedFile->internal.name)) {
			
			//it was the picked file - just unpick it!
			MemPtrFree(mPickedFile);
			mPickedFile = NULL;
		}
	
		MemPtrFree(mCurDetailsFile);
		mCurDetailsFile = t;
	}
	
	return errNone;
}

static Boolean uiPrvSaveExternalDetailsForm(FormPtr fp)
{
	const char *newName = uiPrvGetFieldText(fp, 3000);
	UInt32 attr = 0;
	char *fullPath;
	FileRef f = 0;
	UInt16 i;
	Err e;
	
	for (i = 0; i < 6; i++) {
	
		if (CtlGetValue(uiPrvFrmGetObjectPtrFromId(fp, 2000 + i)))
			attr |= (1UL << i);
	}
	
	fullPath = uiPrvFormVfsName(mCurDetailsFile->name);
	if (!fullPath) {
		e = memErrNotEnoughSpace;
		goto out;
	}
	e = VFSFileOpen(mVolumes[mCurVolume].vrn, fullPath, vfsModeRead, &f);
	if (e != errNone)
		goto out;
	
	e = VFSFileSetAttributes(f, attr &~ (vfsFileAttrDirectory | vfsFileAttrVolumeLabel) /* these MUST be zero */);
	VFSFileClose(f);
	f = 0;
	if (e != errNone)
		goto out;
	
	if (StrCompare(mCurDetailsFile->name, newName)) {

		e = VFSFileRename(mVolumes[mCurVolume].vrn, fullPath, newName);
		if (e != errNone)
			goto out;
		
		e = uiPrvCurFileRenamed(newName);
		if (e != errNone)
			goto out;
		
		StrCopy(mCurDetailsFile->name, newName);
	}
	
out:
	if (fullPath)
		MemPtrFree(fullPath);
	if (f)
		VFSFileClose(f);
	if (e == errNone)
		return true;
		
	uiPrvErrAlert(e);
	return false;
}

static Boolean uiPrvSaveInternalDetailsForm(FormPtr fp)
{
	const char *newName = uiPrvGetFieldText(fp, 3000);
	const char *newType = uiPrvGetFieldText(fp, 3001);
	const char *newCrid = uiPrvGetFieldText(fp, 3002);
	const char *newVerStr = uiPrvGetFieldText(fp, 3003);
	UInt16 newFlags = 0, i, newVer = StrAToI(newVerStr);
	UInt32 type, crid;
	Err e;
	
	if (StrLen(newType) != 4) {
	
		FrmCustomAlert(1000, mStrDatabase, mStrType, mStrMustBe4charsLong);
		return false;
	}
	if (StrLen(newCrid) != 4) {
	
		FrmCustomAlert(1000, mStrDatabase, mStrCrid, mStrMustBe4charsLong);
		return false;
	}
	
	MemMove(&type, newType, 4);
	MemMove(&crid, newCrid, 4);
	
	for (i = 0; i < 12; i++) {
	
		if (CtlGetValue(uiPrvFrmGetObjectPtrFromId(fp, 2000 + i)))
			newFlags |= (1 << i);
	}
	
	e = uiPrvCurFileRenamed(newName);
	if (e != errNone) {
	
		uiPrvErrAlert(e);
		return false;
	}
	
	//DmSetDatabaseInfo will return dmErrAlreadyExists if we attempt to rename a DB to the name it already has. check for that and avoid doing so
	e = DmSetDatabaseInfo(mCurDetailsFile->cardNo, DmFindDatabase(mCurDetailsFile->cardNo, mCurDetailsFile->name), StrCompare(mCurDetailsFile->name, newName) ? newName : NULL, &newFlags, &newVer, NULL, NULL, NULL, NULL, NULL, NULL, &type, &crid);
	if (e != errNone) {
	
		uiPrvErrAlert(e);
		return false;
	}
	
	StrCopy(mCurDetailsFile->name, newName);
	
	return true;
}

static Boolean uiPrvCreateDirectory(FormPtr fp)
{
	char *fullPath;
	Err e;

	fullPath = uiPrvFormVfsName(uiPrvGetFieldText(fp, 3000));
	if (!fullPath)
		return false;
	
	e = VFSDirCreate(mVolumes[mCurVolume].vrn, fullPath);
	MemPtrFree(fullPath);
	
	if (e == errNone)
		return true;
	
	uiPrvErrAlert(e);
	return false;
}

static void uiPrvItemInfo(struct FileEntry *f)
{
	mCurDetailsFile = f;
	
	if (f->type == TYPE_INTERNAL_RAM || f->type == TYPE_INTERNAL_ROM)
		FrmPopupForm(1001);
	else
		FrmPopupForm(1002);
}

static Boolean uiPrvInternalInfoFormEventHandler(EventType *e)
{
	if (e->eType == ctlSelectEvent) switch (e->data.ctlSelect.controlID) {
		
		case 8000:
			if (uiPrvSaveInternalDetailsForm(FrmGetFormPtr(1001))) {
				
				FrmReturnToForm(1000);
				uiPrvFileListResortAndUiRefresh();
			}
			break;
			
		case 8001:
			FrmReturnToForm(1000);
			return true;
	}
	
	return false;
}

static Boolean uiPrvExternalInfoFormEventHandler(EventType *e)
{
	if (e->eType == ctlSelectEvent) switch (e->data.ctlSelect.controlID) {
		
		case 8000:
			if (uiPrvSaveExternalDetailsForm(FrmGetFormPtr(1002))) {
				
				FrmReturnToForm(1000);
				uiPrvFileListResortAndUiRefresh();
			}
			break;
			
		case 8001:
			FrmReturnToForm(1000);
			return true;
	}
	else if (e->eType == EVENT_TYPE_VOLUMES_CHANGED) {		//immediately close it if cards changed, and propagate event
	
		FrmReturnToForm(1000);
		EvtAddEventToQueue(e);
		return true;
	}
	
	return false;
}

static Boolean uiPrvNewFolderNameFormEventHandler(EventType *e)
{
	if (e->eType == ctlSelectEvent) switch (e->data.ctlSelect.controlID) {
		
		case 8000:
			if (uiPrvCreateDirectory(FrmGetFormPtr(1003))) {
				
				FrmReturnToForm(1000);
				uiPrvFileListAndUiRefresh();
			}
			break;
			
		case 8001:
			FrmReturnToForm(1000);
			return true;
	}
	else if (e->eType == EVENT_TYPE_VOLUMES_CHANGED) {		//immediately close it if cards changed, and propagate event
	
		FrmReturnToForm(1000);
		EvtAddEventToQueue(e);
		return true;
	}
	
	return false;
}

static struct FileEntry* uiPrvGetSelectedFile(void)
{
	Int16 selIdx = LstGetSelection(uiPrvFrmGetObjectPtrFromId(FrmGetFormPtr(1000), 1000));
	struct FileEntry *t = NULL;
	
	if (selIdx != noListSelection)
		for (t = mFiles; t && selIdx > 0; selIdx--, t = t->next);
	
	return t;
}

static void uiPrvDirEnter(void)
{
	struct FileEntry *t = uiPrvGetSelectedFile();
	if (!t || t->type != TYPE_VFS_DIR) {
	
		FrmCustomAlert(1000, mStrNeedToSelect, mStrAdir, mStrToPerformThisAction);
	}
	else {

		char *curPath = MemHandleLock(mCurPath);
		UInt32 curLen = StrLen(curPath), newLen = StrLen(t->name);
		Err e;
		
		if (curLen && curPath[curLen - 1] == '/')
			curLen--;
		
		MemHandleUnlock(mCurPath);
		
		e = MemHandleResize(mCurPath, curLen + newLen + 2);
		if (e != errNone) {
		
			uiPrvErrAlert(e);
			return;
		}
		
		curPath = MemHandleLock(mCurPath);
		curPath[curLen++] = '/';
		StrCopy(curPath + curLen, t->name);
		MemHandleUnlock(mCurPath);
		
		uiPrvFileListAndUiRefresh();
	}
}

static void uiPrvDirExit(void)
{
	char *curPath = MemHandleLock(mCurPath);
	Int32 curLen = StrLen(curPath), i;
	
	for (i = curLen - 1; i >= 0 && curPath[i] != '/'; i--);
	
	if (i < 0)	//nothing found?
		return;
	else if (i == 0)	//keep the slash
		i = 1;
	
	curPath[i] = 0;
	MemHandleUnlock(mCurPath);
	
	(void)MemHandleResize(mCurPath, i + 1);
	
	uiPrvFileListAndUiRefresh();
}

static void uiPrvSetDefaultPath(void)
{
	if (!mCurPath)
		mCurPath = MemHandleNew(2);	//if this fails, too much is wrong to care...
	else
		(void)MemHandleResize(mCurPath, 2);
	
	StrCopy(MemHandleLock(mCurPath), mStrDirSeparator);
	MemHandleUnlock(mCurPath);
}

static Boolean uiPrvMainFormEventHandler(EventType *e)
{
	struct FileEntry *t;

	if (e->eType == ctlSelectEvent) switch (e->data.ctlSelect.controlID) {
		case 4000:
			if (mCurVolume < 0) {
						
				mCurVolume = e->data.ctlSelect.on ? VOL_NO_INTERNAL_WITH_ROM : VOL_NO_INTERNAL_NO_ROM;
				
				uiPrvFileListAndUiRefresh();
			}
			break;
		case 6000:
			mCurCompareF = (mCurCompareF == uiPrvSortCompareByName) ? uiPrvSortCompareByNameReverse : uiPrvSortCompareByName;
			uiPrvFileListResortAndUiRefresh();
			break;
		
		case 6001:
			mCurCompareF = (mCurCompareF == uiPrvSortCompareBySize) ? uiPrvSortCompareBySizeReverse : uiPrvSortCompareBySize;
			uiPrvFileListResortAndUiRefresh();
			break;
		
		case 7000:	//enter a folder
			
			uiPrvDirEnter();
			break;
		
		case 7001:	//go up the folder hierarchy
			uiPrvDirExit();
			break;
	}
	else if (e->eType == popSelectEvent) {
		
		Int8 newVolumeIdx;
		
		CtlSetLabel(e->data.popSelect.controlP, uiPrvVolumeListNameByIndex(e->data.popSelect.selection));
		
		newVolumeIdx = e->data.popSelect.selection ? e->data.popSelect.selection - 1 : (CtlGetValue(uiPrvFrmGetObjectPtrFromId(FrmGetFormPtr(1000), 4000)) ? VOL_NO_INTERNAL_WITH_ROM : VOL_NO_INTERNAL_NO_ROM);
		
		if (newVolumeIdx != mCurVolume) {
			mCurVolume = newVolumeIdx;
			uiPrvSetDefaultPath();
			uiPrvFileListAndUiRefresh();
		}
		
		return true;
	}
	else if (e->eType == menuEvent) {
		
		t = uiPrvGetSelectedFile();
		
		//things that do not need an item
		switch (e->data.menu.itemID) {
			case 1004:	//paste
				uiPrvItemPaste();
				return true;
			
			case 2001:	//dir exit
				uiPrvDirExit();
				return true;
				
			case 2002:	//dir create
			
				if (mCurVolume < 0)
					FrmCustomAlert(1000, mStrNeedToSelect, mStrExpansionCard, mStrToPerformThisAction);
				else
					FrmPopupForm(1003);
				return true;
			
			case 9000:	//about
				FrmAlert(1001);
				return true;
		}
		
		if (!t)
			FrmCustomAlert(1000, mStrNeedToSelect, mStrAnItem, mStrToPerformThisAction);
		else switch (e->data.menu.itemID) {

			case 1000:	//info
				uiPrvItemInfo(t);
				break;

			case 1001:	//delete
				uiPrvItemDelete(t);
				break;

			case 1002:	//cut
				uiPrvItemCopy(t, true);
				break;
			
			case 1003:	//copy
				uiPrvItemCopy(t, false);
				break;

			case 1005:	//send
				uiPrvItemSend(t);
				break;
			
			case 2000:	//dir enter
				uiPrvDirEnter();
				break;
		}
	}
	else if (e->eType == EVENT_TYPE_VOLUMES_CHANGED) {
		
		if (uiPrvVolumesChanged()) {
			
			uiPrvSetDefaultPath();
			CtlSetLabel(uiPrvFrmGetObjectPtrFromId(FrmGetActiveForm(), 3000), uiPrvVolumeListNameByIndex(mCurVolume));
			uiPrvFileListAndUiRefresh();
		}
	}
	else if (e->eType == keyDownEvent && (mCurCompareF == uiPrvSortCompareByName || mCurCompareF == uiPrvSortCompareByNameReverse) && !(e->data.keyDown.modifiers & (controlKeyMask | optionKeyMask | commandKeyMask))) {
		
		struct {
			struct FileEntry f;
			char name[3];
		} mFakeEntry;
		UInt16 idx = 0;
		
		mFakeEntry.f.name[0] = e->data.keyDown.chr;
		mFakeEntry.f.name[1] = '0';
		mFakeEntry.f.name[2] = 0;
		
		for (t = mFiles; t && mCurCompareF(&mFakeEntry.f, t) > 0; idx++, t = t->next);
		
		if (t || idx) {
		
			ListPtr list = uiPrvFrmGetObjectPtrFromId(FrmGetFormPtr(1000), 1000);
			
			if (!t)
				idx--;
			
			//this order is important for PalmOS 1.0. Any other order will cause visual corruption in the List
			LstSetSelection(list, -1);
			LstMakeItemVisible(list, idx);
			LstEraseList(list);
			LstDrawList(list);
			LstSetSelection(list, idx);
		}
	}
	return false;
}

static void uiLoop(void)
{
	EventType e;
	FormPtr fp;
	Err err;
	
	while (1) {
		EvtGetEvent(&e, -1);
		
		if (SysHandleEvent(&e))
			continue;
		
		if (MenuHandleEvent(NULL, &e, &err))
			continue;
		
		switch (e.eType) {
			case appStopEvent:
				FrmCloseAllForms();
				return;
			
			case frmLoadEvent:
				fp = FrmInitForm(e.data.frmLoad.formID);
				FrmSetActiveForm(fp);
				
				switch (e.data.frmLoad.formID) {
					case 1000:
						LstSetDrawFunction(uiPrvFrmGetObjectPtrFromId(fp, 1000), uiPrvFileListDrawFunc);
						LstSetDrawFunction(uiPrvFrmGetObjectPtrFromId(fp, 3100), uiPrvVolumeListDrawFunc);
						FrmSetEventHandler(fp, uiPrvMainFormEventHandler);
						CtlSetLabel(uiPrvFrmGetObjectPtrFromId(fp, 3000), uiPrvVolumeListNameByIndex(0));
						CtlSetValue(uiPrvFrmGetObjectPtrFromId(fp, 6000), true);
						uiPrvVolumesChanged();
						uiPrvFileListAndUiRefresh();
						break;
					
					case 1001:
						uiPrvPopulateInternalDetailsForm(fp);
						FrmSetEventHandler(fp, uiPrvInternalInfoFormEventHandler);
						break;
					
					case 1002:
						if (!uiPrvPopulateExternalDetailsForm(fp))
							FrmGotoForm(1000);
						FrmSetEventHandler(fp, uiPrvExternalInfoFormEventHandler);
						break;
					
					case 1003:
						FrmSetEventHandler(fp, uiPrvNewFolderNameFormEventHandler);
						break;
						
				}
				break;
			
			case frmOpenEvent:
				FrmDrawForm(FrmGetActiveForm());
				break;
		}
		
		FrmDispatchEvent(&e);
	}
}

UInt32 __Startup__(void);
UInt32 __Startup__(void)
{
	void *prevGlobalsP, *globalsP;
	SysAppInfoPtr appInfoP;
	
	if (SysAppStartup(&appInfoP, &prevGlobalsP, &globalsP) == errNone) {
		
		if (appInfoP->cmd == sysAppLaunchCmdNormalLaunch) {
		
			UInt16 cardNo;
			UInt32 dummy;
			LocalID LID;
			
			SysCurAppDatabase(&cardNo, &LID);
			
			//subscribe to volume changes, if notif mgr exists
			if (errNone == FtrGet(sysFtrCreator, sysFtrNumNotifyMgrVersion, &dummy) && dummy) {
				
				(void)SysNotifyRegister(cardNo, LID, sysNotifyVolumeMountedEvent, uiPrvVolumesChangedNotif, sysNotifyNormalPriority, NULL);
				(void)SysNotifyRegister(cardNo, LID, sysNotifyVolumeUnmountedEvent, uiPrvVolumesChangedNotif, sysNotifyNormalPriority, NULL);
			}
			
			//init compare func
			mCurCompareF = uiPrvSortCompareByName;
			
			uiPrvSetDefaultPath();
			
			FrmGotoForm(1000);
			uiLoop();

			//unsubscribe to volume changes, if notif mgr exists
			if (errNone == FtrGet(sysFtrCreator, sysFtrNumNotifyMgrVersion, &dummy) && dummy) {
				
				(void)SysNotifyUnregister(cardNo, LID, sysNotifyVolumeMountedEvent, sysNotifyNormalPriority);
				(void)SysNotifyUnregister(cardNo, LID, sysNotifyVolumeUnmountedEvent, sysNotifyNormalPriority);
			}
			
			//cleanup some memory
			if (mCurPath)
				MemHandleFree(mCurPath);
			filesPrvFree(mFiles);
			if (mPickedFile)
				MemPtrFree(mPickedFile);
		}
			
		SysAppExit(appInfoP, prevGlobalsP, globalsP);
	}
	
	return 0;
}





