#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include "swap.h"
#include "frag.h"
#include "slk.h"

#pragma pack (push,1)
PACKED_PRE
struct FragHdr {		//fragmentation is the second-level wrapper for all hotsync protocols. there is another version for laager data XFERs, but fuck it
	uint8_t typ;		//1 = data, 2 = ack
	uint8_t flags;		//0x40 = last, 0x80 = first, 0x10 = long header
	uint16_t ofst;		//for fist fragment: size of total data, BE
}
PACKED_POST
;
#pragma pack (pop)

//acks have 2 bytes of garbage data in them. Ask PalmOS why...


#define FRAG_TYP_DATA			1		//data that will be acked
#define FRAG_TYP_ACK			2		//can have payload but is not fragmentable
#define FRAG_FLAG_LARGE			0x10
#define FRAG_FLAG_OOM			0x20
#define FRAG_FLAG_LAST			0x40
#define FRAG_FLAG_FIRST			0x80


static bool fragPrvRecvFrag(void *slkState, struct SlkHdr *slkHdr, struct FragHdr *fragHdr, struct bv *buf, uint32_t endTime)
{
	struct bv recvVec = BV_INIT(fragHdr, sizeof(struct FragHdr), buf);
	
	//we ignore loopback test packets
	do {
		
		//rx
		if (!slkRx(slkState, slkHdr, &recvVec, endTime))
			return false;
		
	} while (slkHdr->pktTyp == PKT_TYP_LOOPBACK_TEST);
	
	//verify that we got at least enough bytes to make up a fragment header
	if (slkHdr->dataLen < sizeof(struct FragHdr))
		return false;
	
	//it could be a packet...
	return true;
}

static bool fragPrvRecvAck(void *slkState, const struct SlkHdr *sentSlkHdr, const struct FragHdr *sentFragHdr, struct bv *buf, uint32_t endTime)
{
	struct FragHdr gotFragHdr;
	struct SlkHdr gotSlkHdr;
	
	while (fragPrvRecvFrag(slkState, &gotSlkHdr, &gotFragHdr, buf, endTime)) {
	
		//make sure packet type, addresses, and TID match
		if (gotSlkHdr.dstSlk != sentSlkHdr->srcSck || gotSlkHdr.srcSck != sentSlkHdr->dstSlk || gotSlkHdr.pktTyp != sentSlkHdr->pktTyp || gotSlkHdr.tid != sentSlkHdr->tid)
			continue;
		
		//make sure this is an ack/RSP packet and has the proper offset
		if (gotFragHdr.typ != FRAG_TYP_ACK || gotFragHdr.ofst != sentFragHdr->ofst)
			continue;
		
		if (gotFragHdr.flags & FRAG_FLAG_OOM) {
			
			//fprintf(stderr, "Reply claims OOM\n");
			return false;
		}
		
		return true;
	}
	
	return false;
}

static bool fragPrvSendFragWithAckWait(void *slkState, struct SlkHdr *slkHdr, struct FragHdr *fHdr, struct bv *bufTx, struct bv *bufRx)
{
	struct bv sendVec = BV_INIT(fHdr, sizeof(struct FragHdr), bufTx);

	uint_fast8_t i;
	
	for (i = 0; i < FRAG_MAX_RESENDS; i++) {
	
		slkTx(slkState, slkHdr, &sendVec);
		
		if (fragPrvRecvAck(slkState, slkHdr, fHdr, bufRx, slkExtGetCurTime() + FRAG_ACK_WAIT_TIME))
			return true;
	}

	return false;
}

static void fragPrvSendAck(void *slkState, const struct SlkHdr *reqSlkHdr, const struct FragHdr *reqFragHdr, struct bv *bufTx)
{
	struct SlkHdr slk = {.dstSlk = reqSlkHdr->srcSck, .srcSck = reqSlkHdr->dstSlk, .pktTyp = reqSlkHdr->pktTyp, .tid = reqSlkHdr->tid, };
	struct FragHdr fh = {.typ = FRAG_TYP_ACK, .flags = reqFragHdr->flags, .ofst = reqFragHdr->ofst, };
	struct bv sendVec = BV_INIT(&fh, sizeof(struct FragHdr), bufTx);
	
	slkTx(slkState, &slk, &sendVec);
}

bool fragAck(void *slkState, struct SlkHdr *reqSlkHdr, struct FragAckInfo *fi, void *srcP, uint16_t len)
{
	struct FragHdr reconstructedReqFragHdr = {.typ = FRAG_TYP_DATA, .flags = fi->flags, .ofst = fi->ofst, };
	struct bv bv = BV_INIT(srcP, len, NULL);
	
	fragPrvSendAck(slkState, reqSlkHdr, &reconstructedReqFragHdr, len ? &bv : NULL);
	
	return true;
}

bool fragSendStartData(void *slkState, struct FragSendState *fs, struct SlkHdr *slkHdr, uint16_t totalLen)
{
	fs->curPos = 0;
	fs->slkHdr = *slkHdr;
	fs->totalLen = totalLen;
	
	return true;
}

bool fragSendContinue(void *slkState, struct FragSendState *fs, const void *srcP, uint16_t len, struct bv *bufAckRx)
{
	uint8_t *src = (uint8_t*)srcP;
	uint16_t now, ofst;
	
	//zero-length segments mean nothing
	if (!len)
		return true;
	
	//verify we have space
	if (fs->totalLen - fs->curPos < len)
		return false;
	
	//send one or more chunks
	for (ofst = 0; ofst < len; ofst += now, fs->curPos += now) {
		
		struct FragHdr fHdr = {.typ = FRAG_TYP_DATA, };
		struct bv bv = BV_INIT(src + ofst, 0, NULL);
		
		now = len - ofst;
		if (now > FRAG_MAX_CHUNK_SZ)
			now = FRAG_MAX_CHUNK_SZ;
		
		if (fs->curPos)
			fHdr.ofst = TO_U16(fs->curPos);
		else {
			fHdr.ofst = TO_U16(fs->totalLen);
			fHdr.flags |= FRAG_FLAG_FIRST;
		}
		
		if (fs->curPos + now == fs->totalLen)
			fHdr.flags |= FRAG_FLAG_LAST;
		
		//fprintf(stderr, "  sending frag with %u bytes\n", now);
		
		BV_SET_LEN(&bv, now);
		
		if (!fragPrvSendFragWithAckWait(slkState, &fs->slkHdr, &fHdr, &bv, bufAckRx))
			return false;
	}
	
	return true;
}

bool fragSend(void *slkState, struct SlkHdr *slkHdr, const void *srcP, uint16_t len, struct bv *bufAckRx)
{
	struct FragSendState fs;
	
	return fragSendStartData(slkState, &fs, slkHdr, len) && fragSendContinue(slkState, &fs, srcP, len, bufAckRx);
}

bool fragRecv(void *slkState, struct SlkHdr *slkHdr, void *dstP, uint16_t *lenP /* in / out */, uint32_t endTime, bool autoAck, struct FragAckInfo *fi)
{
	uint32_t len = *lenP, done = 0, expected;
	uint8_t *dst = (uint8_t*)dstP;
	bool waitForFirst = true;
	
	while(1) {
		
		struct FragHdr nowFragHdr;
		struct SlkHdr nowSlkHdr;
		uint16_t nowBytes, nowOfst;
		struct bv bv = BV_INIT(dst, len - done, NULL);
		
		if (!fragPrvRecvFrag(slkState, &nowSlkHdr, &nowFragHdr, &bv, endTime))
			return false;
		
		nowBytes = nowSlkHdr.dataLen - sizeof(struct FragHdr);
		
		//only accept data packets
		if (nowFragHdr.typ != FRAG_TYP_DATA)
			continue;
		
		//if we're waiting for the first packet and this is not one, ignore it
		if (waitForFirst && !(nowFragHdr.flags & FRAG_FLAG_FIRST))
			continue;
		else if (waitForFirst) {	//first packet
			
			expected = TO_U16(nowFragHdr.ofst);
			nowOfst = 0;
			waitForFirst = false;
			
			//reject if it is too long
			if (expected > len)
				return false;
			
			*slkHdr = nowSlkHdr;
		}
		else 
			nowOfst = TO_U16(nowFragHdr.ofst);
		
		//verify packet addressing and type is same as first was (ignoreable if not)
		if (nowSlkHdr.dstSlk != slkHdr->dstSlk || nowSlkHdr.srcSck != slkHdr->srcSck || nowSlkHdr.pktTyp != slkHdr->pktTyp || nowSlkHdr.tid != slkHdr->tid)
			continue;
		
		//verify the offset if as expected (out-of-order packet is ignoreable)
		if (nowOfst != done)
			continue;
		
		//verify we did not overrun the end (mistake by sender - failure)
		if (expected - done < nowBytes)
			return false;
		
		//verify end packet properly ends, if this is the end (mistake by sender - failure)
		if (!(nowFragHdr.flags & FRAG_FLAG_LAST) != !(expected - done == nowBytes))
			return false;
		
		//if we got this far, we're accepting this fragment - ACK it (except for last packet in a series if auto-ack is off
		if (autoAck || !(nowFragHdr.flags & FRAG_FLAG_LAST))
			fragPrvSendAck(slkState, &nowSlkHdr, &nowFragHdr, NULL);
		
		if (fi) {
			fi->ofst = nowFragHdr.ofst;
			fi->flags = nowFragHdr.flags;
		}
		
		//account for it
		done += nowBytes;
		
		if (expected == done) {
			*lenP = done;
			return true;
		}
	}		
}




